/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.curves;

import java.util.ArrayList;
import java.util.List;

import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 *
 */
public class ArrivalCurve extends Curve {
	@SuppressWarnings("unused")
	private ArrivalCurve() {
		super();
	}
	
	public ArrivalCurve( int segment_count ) {
		super( segment_count );
	}
		
	public ArrivalCurve( Curve curve ) {
		super();

		if( Configuration.performArrivalCurveChecks() && !this.isWideSenseIncreasing() ) {
			throw new RuntimeException("Arrival curves can only be created from wide sense increasing curves.");
		}
		
		alias = curve.getAlias();
		label = curve.getLabel();

		token_buckets = curve.token_buckets;
		rate_latencies = curve.rate_latencies;

		int i = 0;
		if ( curve.getSegment(0).y.greater( Num.ZERO ) ) {
			segments = new LinearSegment[curve.getSegmentCount()+1];
			
			segments[0] = LinearSegment.createNullSegment();
			i++;
			
			segments[1] = curve.getSegment( 0 ).copy();
			segments[1].leftopen = true;
			i++;
		} else {
			segments = new LinearSegment[curve.getSegmentCount()];
		}
		
		for ( ; i < curve.getSegmentCount(); i++ ) {
			segments[i] = curve.getSegment( i ).copy();
		}
	}
	
	/**
	 * Creates a null arrival curve.
	 * 
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createNullArrival() {
		return ( new ArrivalCurve( Curve.NULL_CURVE ) ).copy();
	}
	
	/**
	 * Creates a burst curve with zero delay.
	 * 
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createZeroDelayBurst() {
		return new ArrivalCurve( ZERO_DELAY_BURST_CURVE );
	}

	/**
	 * Creates a new token bucket curve.
	 * 
	 * @param r the rate
	 * @param b the burstiness
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createTokenBucket( Num r, Num b ) {
		ArrivalCurve result = new ArrivalCurve( Curve.createTokenBucket(r, b) );
		result.token_buckets = new ArrayList<Curve>();
		result.token_buckets.add( result.copy() );
		result.is_token_bucket = true;
		
		return result;
	}

	/**
	 * Creates a new token bucket curve.
	 * 
	 * @param r the rate
	 * @param b the burstiness
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createTokenBucket( double r, double b ) {
		ArrivalCurve result = new ArrivalCurve( Curve.createTokenBucket(r, b) );
		result.token_buckets = new ArrayList<Curve>();
		result.token_buckets.add( result.copy() );
		result.is_token_bucket = true;
		
		return result;
	}

	/**
	 * Creates a new curve from a list of token bucket curves.
	 * 
	 * @param token_buckets a list of token bucket curves
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createFromTokenBuckets( List<Curve> token_buckets ) {
		return new ArrivalCurve( Curve.createFromTokenBuckets( token_buckets ) );
	}
	
	public Num getBurst() {
		return fLimitRight( Num.ZERO );
	}
	
	public ArrivalCurve copy() {
		return new ArrivalCurve( this );
	}

	public static ArrivalCurve add( ArrivalCurve c1, ArrivalCurve c2 ) {
		return new ArrivalCurve( Curve.add( c1, c2 ) );
	}

	public static ArrivalCurve add( ArrivalCurve c1, Num dy ) {
		return new ArrivalCurve( Curve.add( c1, dy ) );
	}
	
	/**
	 * Returns a copy of this curve that is shifted to the right by <code>dx</code>,
	 * i.e. g(x) = f(x-dx).
	 * 
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public ArrivalCurve shiftRight( Num dx ) {
		return new ArrivalCurve( Curve.shiftRight( this, dx ) );
	}
	
	/**
	 * Returns a copy of this curve that is shifted to the left by <code>dx</code>,
	 * i.e. g(x) = f(x+dx). Note that the new curve is clipped at the y-axis so
	 * that in most cases <code>c.shiftLeftClipping(dx).shiftRight(dx) != c</code>!
	 * 
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public ArrivalCurve shiftLeftClipping( Num dx ) {
		return new ArrivalCurve( Curve.shiftLeftClipping( this, dx ) );
	}
	
	/**
	 * Returns a string representation of this curve.
	 * 
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		return "AC" + super.toString();
	}
}
