/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.curves;

import java.util.ArrayList;

import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 *
 */
public class ServiceCurve extends Curve {
	public ServiceCurve() {
		super();
	}

	public ServiceCurve( int segment_count ) {
		super( segment_count );
	}
	
	public ServiceCurve( Curve curve ) {
		super( curve.getSegmentCount() );
		
		if( Configuration.performServiceCurveChecks() && !this.isWideSenseIncreasing() ) {
			throw new RuntimeException("Service curves can only be created from wide sense increasing curves.");
		}
		
		alias = curve.getAlias();
		label = curve.getLabel();

		token_buckets = curve.token_buckets;
		rate_latencies = curve.rate_latencies;
		
		for ( int i = 0; i < curve.getSegmentCount(); i++ ) {
			segments[i] = curve.getSegment( i ).copy();
		}
	}
	
	/**
	 * Creates a null arrival curve.
	 * 
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createNullService() {
		return ( new ServiceCurve( Curve.NULL_CURVE ) ).copy();
	}
	
	/**
	 * Creates a burst curve with zero delay.
	 * 
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createZeroDelayBurst() {
		return new ServiceCurve( ZERO_DELAY_BURST_CURVE );
	}

	/**
	 * Creates a burst delay curve.
	 * 
	 * @param t the delay, which must be >= 0.0
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createBurstDelay( Num t ) {
		return new ServiceCurve( Curve.createBurstDelay( t ) );
	}
	
	/**
	 * Creates a new rate latency curve.
	 * 
	 * @param r the rate
	 * @param t the latency
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createRateLatency( Num r, Num t ) {
		ServiceCurve result = new ServiceCurve( Curve.createRateLatency( r, t ) );
		result.rate_latencies = new ArrayList<Curve>();
		result.rate_latencies.add( result.copy() );
		result.is_rate_latency = new Boolean( true );
		
		return result;
	}

	/**
	 * Creates a new rate latency curve.
	 * 
	 * @param r the rate
	 * @param t the latency
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createRateLatency( double r, double t ) {
		ServiceCurve result = new ServiceCurve( Curve.createRateLatency( new Num( r ), new Num( t ) ) );
		result.rate_latencies = new ArrayList<Curve>();
		result.rate_latencies.add( result.copy() );
		result.is_rate_latency = new Boolean( true );
		
		return result;
	}

	public ServiceCurve copy() {
		return new ServiceCurve( this );
	}
	
	public static ServiceCurve add( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.add(c1, c2) );
	}
	
	public static ServiceCurve sub( ServiceCurve c1, ArrivalCurve c2 ) {
		return new ServiceCurve( Curve.sub(c1, c2) );
	}
	
	public static ServiceCurve min( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.min(c1, c2) );
	}
	
	public static ServiceCurve max( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.max(c1, c2) );
	}
	
	/**
	 * Returns a copy of curve bounded at the x-axis.
	 * 
	 * @return the bounded curve.
	 */
	public static ServiceCurve boundAtXAxis( ServiceCurve service_curve ) {
		return new ServiceCurve( Curve.boundAtXAxis( service_curve ) ); 
	}
	
	/**
	 * Returns a copy of this curve with latency removed, i.e. shifted left by
	 * the latency.
	 * 
	 * @return a copy of this curve without latency
	 */
	public ServiceCurve removeLatency() {
		return new ServiceCurve( Curve.removeLatency( this ) );
	}
	
	public boolean equals( ServiceCurve other ) {
		return ( (Curve)this ).equals( other );
	}
	
	/**
	 * Returns a string representation of this curve.
	 * 
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		return "SC" + super.toString();
	}
}
