/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.misc;

import org.apache.commons.math3.fraction.Fraction;

/**
 * Wrapper class around org.apache.commons.math3.fraction.Fraction
 * introducing special values like positive / negative infinity and NaN
 * as well as operators like min, max, ==, >, >=, <, and >= that are
 * not part of Fraction but needed by the network calculator.
 * 
 * For the ease of converting from the primitive data type double
 * to Fraction objects, copy by value semantic are is applied. 
 * 
 * @author Steffen Bondorf
 *
 */
public class Num extends Fraction {
	public static enum RepresentationFlag { FRACTION, DOUBLE };
	public static final RepresentationFlag representation = RepresentationFlag.FRACTION;
	
	private static final long serialVersionUID = 1L;
	
	// Fraction is based in Integer and thus there's no infinity (and Fraction is prone to overflows as well)
	public static final Num POSITIVE_INFINITY = new Num( Integer.MAX_VALUE );
	public static final Num NEGATIVE_INFINITY = new Num( Integer.MIN_VALUE );
	
	public static final Num NaN = new Num( Double.NaN );
	
	public static final Num ZERO = new Num( 0.0 );
	
	public Num( double value ) {
		super( value );
	}
	
	public Num( int num ) {
		super( num );
	}
	
	public Num( int num, int den ) {
		super( num, den );
	}
	
	public Num( Num num ) {
		super( num.getNumerator(), num.getDenominator() );
	}
	
	private Num( Fraction frac ) {
		super( frac.getNumerator(), frac.getDenominator() );
	}
	
	public static Num getEpsilon() {
        // Unfortunately you cannot give the constructor the double value 0.0000001
        return new Num( 1, 1000000 );
	}

	// In order to simplify the transition from the primitive data type double to
	// a rational number object these functions emulate copy by value for objects that
	// typically inhibit copy by reference
	public static Num add( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		// Prevent overflow exception when adding integer based number representations like Fraction
		if( num1 == POSITIVE_INFINITY || num2 == POSITIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}
		if( num1 == NEGATIVE_INFINITY || num2 == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
		
		// May throw MathArithmeticException due to integer overflow
        return new Num( num1.add( num2 ) );
	}
	
	public static Num sub( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		// Prevent overflow exception when adding integer based number representations like Fraction
		if( num1 == POSITIVE_INFINITY || num2 == POSITIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}
		if( num1 == NEGATIVE_INFINITY || num2 == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
		
        // May throw MathArithmeticException due to integer overflow
        return new Num( num1.subtract( num2 ) );
	}
	
	public static Num mult( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		// Prevent overflow exception when adding integer based number representations like Fraction
		if( num1 == POSITIVE_INFINITY || num2 == POSITIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}
		if( num1 == NEGATIVE_INFINITY || num2 == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
		
        // May throw MathArithmeticException due to integer overflow
       	return new Num( num1.multiply( num2 ) );
	}

	public static Num div( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		
		// Integer based number representations use Integer.MAX_VALUE to signal infinity so special treatment is necessary when dividing
		if( num1 == POSITIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}
		if( num2 == POSITIVE_INFINITY ) {
			return ZERO;
		}
		if( num1 == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
		if( num2 == NEGATIVE_INFINITY ) {
			return ZERO;
		}
		
        if ( num2.getNumerator() == 0 ) {
        	return Num.POSITIVE_INFINITY;
       	} else {
           	return new Num( num1.divide( num2 ) );        		
       	}
	}

	public static Num diff( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		
		// Integer based number representations use Integer.MAX_VALUE to signal infinity so special treatment is necessary when dividing
		if( num1 == POSITIVE_INFINITY || num1 == NEGATIVE_INFINITY 
				 || num2 == POSITIVE_INFINITY || num2 == NEGATIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}
		
		return sub( max( num1, num2 ), min( num1, num2 ) );	
	}

	public Num copy() {
		if ( this == NaN ) {
    		return NaN;
    	}
    	if ( this == POSITIVE_INFINITY ) {
    		return POSITIVE_INFINITY;
    	}
    	if ( this == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
    	
		return new Num( this.getNumerator(), this.getDenominator() );
	}

	public static Num max( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		if( num1 == POSITIVE_INFINITY ) {
			return num1;
		}
		if( num1 == NEGATIVE_INFINITY ) {
			return num2;
		}
		
		// Operate on Doubles to prevent getting out of Integer's range    	
        double numerator1 = new Double( num1.getNumerator() ) * new Double( num2.getDenominator() );
        double numerator2 = new Double( num2.getNumerator() ) * new Double( num1.getDenominator() );
        	
        if ( numerator1 >= numerator2 ) {
        	return num1;
        } else {
        	return num2;
        }
	}

	public static Num min( Num num1, Num num2 ) {
		if( num1 == NaN || num2 == NaN ) {
			return NaN;
		}
		if( num1 == POSITIVE_INFINITY ) {
			return num2;
		}
		if( num1 == NEGATIVE_INFINITY ) {
			return num1;
		}
		
        // Operate on Doubles to prevent getting out of Integer's range 
        double numerator1 = new Double( num1.getNumerator() ) * new Double( num2.getDenominator() );
        double numerator2 = new Double( num2.getNumerator() ) * new Double( num1.getDenominator() );
        	
       	if ( numerator1 < numerator2 ) {
       		return num1;
        } else {
        	return num2;
        }
	}
	
	public boolean equals( double num2 ) {
		if( num2 == Double.NaN ){
			return this == NaN;
		}
		if( num2 == Double.POSITIVE_INFINITY ){
			return this == POSITIVE_INFINITY;
		}
		if( num2 == Double.NEGATIVE_INFINITY ){
			return this == NEGATIVE_INFINITY;
		}
		return equals( new Num( num2 ) );
	}

	public boolean equals( Num num2 ) {
		if( this == NaN & num2 == NaN ){
			return true;
		}
		if( this == POSITIVE_INFINITY & num2 == POSITIVE_INFINITY ){
			return true;
		}
		if( this == NEGATIVE_INFINITY & num2 == NEGATIVE_INFINITY ){
			return true;
		}
        // Operate on Doubles to prevent getting out of Integer's range 
        double numerator1 = new Double( this.getNumerator() ) * new Double( num2.getDenominator() );
        double numerator2 = new Double( num2.getNumerator() ) * new Double( this.getDenominator() );
        	
        if ( numerator1 == numerator2 ) {
        	return true;
        } else {
        	return false;
        }
	}

	@Override
	public boolean equals( Object num2 ) {
		Num num2_Num;
		try {
			num2_Num = (Num) num2;
			return equals( num2_Num );
		} catch( ClassCastException e ) {
			return false;
		}
	}

	public boolean greater( Num num2 ) {
		if( this == NaN || num2 == NaN ){
			return false;
		}
		
		if( num2 == POSITIVE_INFINITY ){
			return false;
		}
		if( this == POSITIVE_INFINITY ){
			return true;
		}
		
		if( this == NEGATIVE_INFINITY ){
			return false;
		}
		if( num2 == NEGATIVE_INFINITY ){
			return true;
		}
		
        // Operate on Doubles to prevent getting out of Integer's range 
        double numerator1 = new Double( this.getNumerator() ) * new Double( num2.getDenominator() );
       	double numerator2 = new Double( num2.getNumerator() ) * new Double( this.getDenominator() );
     	
       	if ( numerator1 > numerator2 ) {
       		return true;
       	} else {
       		return false;
       	}
	}

	public boolean ge( Num num2 ) {
		if( this == NaN || num2 == NaN ){
			return false;
		}
		
		if( this == POSITIVE_INFINITY ){
			return true;
		}
		if( num2 == POSITIVE_INFINITY ){
			return false;
		}

		if( num2 == NEGATIVE_INFINITY ){
			return true;
		}
		if( this == NEGATIVE_INFINITY ){
			return false;
		}
		
		// Operate on Doubles to prevent getting out of Integer's range 
        double numerator1 = new Double( this.getNumerator() ) * new Double( num2.getDenominator() );
        double numerator2 = new Double( num2.getNumerator() ) * new Double( this.getDenominator() );
        	
        if ( numerator1 >= numerator2 ) {
        	return true;
        } else {
        	return false;
        }
	}

	public boolean less( Num num2 ) {
		if( this == NaN || num2 == NaN ){
			return false;
		}

		if( this == POSITIVE_INFINITY ){
			return false;
		}
		if( num2 == POSITIVE_INFINITY ){
			return true;
		}
		
		if( num2 == NEGATIVE_INFINITY ){
			return false;
		}
		if( this == NEGATIVE_INFINITY ){
			return true;
		}
			
       	// Operate on Doubles to prevent getting out of Integer's range 
       	double numerator1 = new Double( this.getNumerator() ) * new Double( num2.getDenominator() );
       	double numerator2 = new Double( num2.getNumerator() ) * new Double( this.getDenominator() );
       	
       	if ( numerator1 < numerator2 ) {
       		return true;
       	} else {
       		return false;
       	}
	}

	public boolean le( Num num2 ) {
		if( this == NaN || num2 == NaN ){
			return false;
		}

		if( num2 == POSITIVE_INFINITY ){
			return true;
		}
		if( this == POSITIVE_INFINITY ){
			return false;
		}

		if( this == NEGATIVE_INFINITY ){
			return true;
		}
		if( num2 == NEGATIVE_INFINITY ){
			return false;
		}
		
        // Operate on Doubles to prevent getting out of Integer's range 
       	double numerator1 = new Double( this.getNumerator() ) * new Double( num2.getDenominator() );
       	double numerator2 = new Double( num2.getNumerator() ) * new Double( this.getDenominator() );
       	
       	if ( numerator1 <= numerator2 ) {
       		return true;
       	} else {
       		return false;
       	}
	}
	
	public static Num abs( Num num ) {
		if ( num == NaN ) {
    		return NaN;
    	}
    	if ( num == POSITIVE_INFINITY ) {
    		return POSITIVE_INFINITY;
    	}
    	if ( num == NEGATIVE_INFINITY ) {
			return NEGATIVE_INFINITY;
		}
	    
	    if( num.le( ZERO ) ){
	    	return negate( num );
	    }
	    else
	    {
	    	return num.copy();
	    }
	}

	public static Num negate( Num num ) {
		if ( num == NaN ) {
    		return NaN;
    	}
    	if ( num == POSITIVE_INFINITY ) {
    		return NEGATIVE_INFINITY;
    	}
    	if ( num == NEGATIVE_INFINITY ) {
			return POSITIVE_INFINITY;
		}

	    return Num.mult( num, new Num( -1.0 ) );
	}
	
	@Override
	public double doubleValue() {
		if ( this == NaN ) {
    		return Double.NaN;
    	}
    	if ( this == POSITIVE_INFINITY ) {
    		return Double.POSITIVE_INFINITY;
    	}
    	if ( this == NEGATIVE_INFINITY ) {
			return Double.NEGATIVE_INFINITY;
		}
	    return super.doubleValue();
	}

	@Override
	public float floatValue() {
		if ( this == NaN ) {
    		return Float.NaN;
    	}
    	if ( this == POSITIVE_INFINITY ) {
    		return Float.POSITIVE_INFINITY;
    	}
    	if ( this == NEGATIVE_INFINITY ) {
			return Float.NEGATIVE_INFINITY;
		}
	    return super.floatValue();
	}
	
	@Override
	public String toString(){
		if ( this == NaN ) {
    		return "NaN";
    	}
    	if ( this == POSITIVE_INFINITY ) {
    		return "Infinity";
    	}
    	if ( this == NEGATIVE_INFINITY ) {
			return "-Infinity";
		}
    	
		return super.toString();
	}
}
