/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.minplus.Deconvolution;
import unikl.disco.misc.SetUtils;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;
import unikl.disco.network.Link;

/**
 * 
 * @author Steffen Bondorf
 *
 */
public class ArrivalBound {
	private Network network;
	private Set<Flow> flows_to_bound;
	private Flow flow_of_interest;
	
	private Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>();
	
	protected ArrivalBound() {}
	
	private ArrivalBound( Network network, Server server, Set<Flow> flows_to_bound, Flow flow_of_interest ) {
		this.network = network;
		this.flows_to_bound = flows_to_bound;
		this.flow_of_interest = flow_of_interest;
	}
	
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Server server ) throws Exception {
		return computeArrivalBound( network, server, network.getFlows( server ), Flow.NULL_FLOW );
	}
	
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Server server, Set<Flow> flows_to_bound ) throws Exception {
		return computeArrivalBound( network, server, flows_to_bound, Flow.NULL_FLOW );
	}
	
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Link link ) throws Exception {
		return computeArrivalBound( network, link.getDest(), network.getFlows( link ), Flow.NULL_FLOW );
	}
	
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Link link, Set<Flow> flows_to_bound ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( flows_to_bound, network.getFlows( link ) );
		return computeArrivalBound( network, link.getDest(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}
	
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Link link, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( flows_to_bound, network.getFlows( link ) );
		return computeArrivalBound( network, link.getDest(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}
	
	/**
	 * Calculates the alternative arrival curves for a the flows in flows_to_bound that arrive at the given server
	 * while considering the lower priority of the flow_of_interest
	 * 
	 * @param network
	 * @param server
	 * @param flows_to_bound
	 * @return the arrival bound at the specified server
	 * @throws Exception
	 */
	public static Set<ArrivalCurve> computeArrivalBound( Network network, Server server, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception {
		flows_to_bound.remove( flow_of_interest );
		
		Set<ArrivalCurve> arrival_bounds = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		
		if ( flows_to_bound.size() == 0 )
		{
			return arrival_bounds;
		}
		
		Set<Flow> f_server = network.getFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, flows_to_bound );
		if ( f_xfcaller_server.size() == 0 )
		{
			return arrival_bounds;
		}
		
		ArrivalBound arrival_bounding;
		
		Set<ArrivalCurve> arrival_bounds_link = new HashSet<ArrivalCurve>();
		Set<ArrivalCurve> arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
	
		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		f_xfcaller_sourceflows_server.remove( flow_of_interest );
		ArrivalCurve alpha_xfcaller_sourceflows_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );
		arrival_bounds = new HashSet<ArrivalCurve>( Collections.singleton( alpha_xfcaller_sourceflows_server ) );
		
		// Get cross-traffic from each predecessor. Call per link in order to get splitting points
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		if ( !f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
			for ( Link in_l : network.getInLinks( server ) ) {
				arrival_bounds_link = new HashSet<ArrivalCurve>();
				arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
				
				Set<Flow> f_xfcaller_in_l = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
				f_xfcaller_in_l.remove( flow_of_interest );
				
				if( f_xfcaller_in_l.isEmpty() ) { // Do not check links without flows of interest
					continue;
				}
				
				arrival_bounding = new ArrivalBound( network, server, f_xfcaller_in_l, flow_of_interest );
				arrival_bounds_link = arrival_bounding.computeArrivalBound( in_l );

				if ( !arrival_bounds.isEmpty() ) {
					for ( ArrivalCurve arrival_bound_other_link : arrival_bounds ) {
						for ( ArrivalCurve arrival_bound_link : arrival_bounds_link ) {
							arrival_bound_link.beautify();
							arrival_bounds_link_merged.add( ArrivalCurve.add( arrival_bound_other_link, arrival_bound_link ) );
						}
					}
				} else {
					arrival_bounds_link_merged.addAll( arrival_bounds_link );
				}
				arrival_bounds = arrival_bounds_link_merged;
				
				f_xfcaller_server_bounded.addAll( f_xfcaller_in_l ); // Stop as soon as all flows are bounded
				if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
					break;
				}
			}
		}
		
		return arrival_bounds;
	}

	private Set<ArrivalCurve> computeArrivalBound( Link link ) throws Exception {		
		for( Configuration.ArrivalBoundMethods arrival_bound_method : Configuration.arrivalBoundMethods() ) {
			
			Set<ArrivalCurve> arrival_bounds_tmp = new HashSet<ArrivalCurve>();

			// Variables for per flow bounds
			Set<ArrivalCurve> arrival_bounds_f = new HashSet<ArrivalCurve>();
			Set<ArrivalCurve> arrival_bounds_merged = new HashSet<ArrivalCurve>();
			
			switch( arrival_bound_method ) {
				case PBOO_OUTPUT_AC:
					PbooArrivalBound_Output_PerHop pboo_output_ac_based = new PbooArrivalBound_Output_PerHop( network, flow_of_interest );
					ArrivalCurve arrival_bound_tmp = pboo_output_ac_based.computeArrivalBound( link, flows_to_bound );
					arrival_bound_tmp.beautify();
	
					addArrivalBound( arrival_bound_tmp );
				break;
			
				case PBOO_PER_HOP:
				default:
					PbooArrivalBound_PerHop pboo_per_hop = new PbooArrivalBound_PerHop( network, flow_of_interest );
					arrival_bounds_tmp = pboo_per_hop.computeArrivalBound( link, flows_to_bound );
	
					addArrivalBounds( arrival_bounds_tmp );
				break;
				
				case PBOO_CONCATENATION:
					PbooArrivalBound_Concatenation pboo_concatenation = new PbooArrivalBound_Concatenation( network, flow_of_interest );
					arrival_bounds_tmp = pboo_concatenation.computeArrivalBound( link, flows_to_bound );
					
					addArrivalBounds( arrival_bounds_tmp );
				break;
	
				case PMOO:
					PmooArrivalBound pmoo_arrival_bound = new PmooArrivalBound( network, flow_of_interest );
					arrival_bounds_tmp = pmoo_arrival_bound.computeArrivalBound( link, flows_to_bound );
					
					addArrivalBounds( arrival_bounds_tmp );
				break;
				
				// Functional tests missing
				case PER_FLOW_SFA:					
					for ( Flow flow : flows_to_bound ) {
						arrival_bounds_f = new HashSet<ArrivalCurve>();
						arrival_bounds_merged = new HashSet<ArrivalCurve>();
						
						SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );
						sfa.performAnalysisOnPath( flow, flow.getSubPath( flow.getSource(), link.getDest() ) );
						
						// All permutations of single flow results
						for( ServiceCurve beta_lo : sfa.getLeftOverServiceCurves() ) {
							arrival_bounds_f.add( Deconvolution.deconvolve( flow.getArrivalCurve(), beta_lo ) );
						}
						
						if ( !arrival_bounds_tmp.isEmpty() ) {
							for ( ArrivalCurve alpha : arrival_bounds_tmp ) {
								for ( ArrivalCurve alpha_f : arrival_bounds_f ) {
									alpha_f.beautify();
									arrival_bounds_merged.add( ArrivalCurve.add( alpha, alpha_f ) );
								}
							}
						} else {
							arrival_bounds_merged.addAll( arrival_bounds_f );
						}
						arrival_bounds_tmp = arrival_bounds_merged;
					}
					addArrivalBounds( arrival_bounds_tmp );
				break;
	
				// Functional tests missing
				case PER_FLOW_PMOO:
					for ( Flow flow : flows_to_bound ) {
						arrival_bounds_f = new HashSet<ArrivalCurve>();
						arrival_bounds_merged = new HashSet<ArrivalCurve>();
						
						PmooAnalysis pmoo = new PmooAnalysis( network );
						pmoo.performAnalysisOnPath( flow, flow.getSubPath( flow.getSource(), link.getDest() ) );
			
						// All permutations of single flow results
						for( ServiceCurve beta_lo : pmoo.getLeftOverServiceCurves() ) {
							arrival_bounds_f.add( Deconvolution.deconvolve( flow.getArrivalCurve(), beta_lo ) );
						}
						
						if ( !arrival_bounds_tmp.isEmpty() ) {
							for ( ArrivalCurve alpha : arrival_bounds_tmp ) {
								for ( ArrivalCurve alpha_f : arrival_bounds_f ) {
									alpha_f.beautify();
									arrival_bounds_merged.add( ArrivalCurve.add( alpha, alpha_f ) );
								}
							}
						} else {
							arrival_bounds_merged.addAll( arrival_bounds_f );
						}
						arrival_bounds_tmp = arrival_bounds_merged;
					}
					addArrivalBounds( arrival_bounds_tmp );
				break;
			}
		}
		
		return alphas_xfcaller;
	}
	
	private void addArrivalBounds( Set<ArrivalCurve> arrival_bounds ) {
		for( ArrivalCurve arrival_bound : arrival_bounds ) {
			addArrivalBound( arrival_bound );
		}
	}
	
	private void addArrivalBound( ArrivalCurve arrival_bound ) {
		if( !Configuration.removeDuplicateArrivalBounds() 
				|| ( Configuration.removeDuplicateArrivalBounds() && !isDuplicate( arrival_bound )  ) ) {
			alphas_xfcaller.add( arrival_bound );
		}
	}
	
	private boolean isDuplicate( ArrivalCurve arrival_bound ) {
		for ( ArrivalCurve arrival_bound_existing : alphas_xfcaller ) {
			if( arrival_bound.equals( arrival_bound_existing ) ) {
				return true;
			}
		}
		return false;
	}
}