/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

/**
 * 
 * @author Steffen Bondorf
 *
 */
public class PbooArrivalBound_PerHop extends ArrivalBound {
	Network network;
	Flow flow_of_interest;
	 
	@SuppressWarnings("unused")
	private PbooArrivalBound_PerHop() {}

	public PbooArrivalBound_PerHop( Network network, Flow flow_of_interest ) {
		this.network = network;
		this.flow_of_interest = flow_of_interest;
	}

	public Set<ArrivalCurve> computeArrivalBound( Server server ) throws Exception {
		return computeArrivalBound( server, network.getFlows( server ) );
	}

	public Set<ArrivalCurve> computeArrivalBound( Link link ) throws Exception {
		return computeArrivalBound( link, network.getFlows( link ) );
	}
	
	public Set<ArrivalCurve> computeArrivalBound( Server server, Set<Flow> f_xfcaller ) throws Exception {
		Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		
		f_xfcaller.remove( flow_of_interest );
		if ( f_xfcaller.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		Set<Flow> f_server = network.getFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		if ( f_xfcaller_server.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		Set<ArrivalCurve> arrival_bounds_link = new HashSet<ArrivalCurve>();
		Set<ArrivalCurve> arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
	
		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		f_xfcaller_sourceflows_server.remove( flow_of_interest );
		ArrivalCurve alpha_xfcaller_sourceflows_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );
		alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( alpha_xfcaller_sourceflows_server ) );
		
		// Get cross-traffic from each predecessor. Call per link in order to get splitting points
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		if ( !f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
			for ( Link in_l : network.getInLinks( server ) ) {
				arrival_bounds_link = new HashSet<ArrivalCurve>();
				arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
				
				Set<Flow> f_xfcaller_in_l = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
				f_xfcaller_in_l.remove( flow_of_interest );
				
				if( f_xfcaller_in_l.isEmpty() ) { // Do not check links without flows to bound
					continue;
				}
				
				arrival_bounds_link = computeArrivalBound( in_l, f_xfcaller_in_l );
				
				if ( !alphas_xfcaller.isEmpty() ) {
					for ( ArrivalCurve arrival_bound_other_link : alphas_xfcaller ) {
						for ( ArrivalCurve arrival_bound_link : arrival_bounds_link ) {
							arrival_bound_link.beautify();
							arrival_bounds_link_merged.add( ArrivalCurve.add( arrival_bound_other_link, arrival_bound_link ) );
						}
					}
				} else {
					arrival_bounds_link_merged.addAll( arrival_bounds_link );
				}
				alphas_xfcaller = arrival_bounds_link_merged;
				
				f_xfcaller_server_bounded.addAll( f_xfcaller_in_l ); // Stop as soon as all flows are bounded
				if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
					break;
				}
			}
		}
		
		return alphas_xfcaller;
	}
	
	public Set<ArrivalCurve> computeArrivalBound( Link link, Set<Flow> f_xfcaller ) throws Exception {
		Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		if ( f_xfcaller.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		// Get the servers on common sub-path of f_xfcaller flows crossing link
		Server to = link.getDest();
		Set<Flow> f_to = network.getFlows( to );
		Set<Flow> f_xfcaller_to = SetUtils.getIntersection( f_to, f_xfcaller );
		f_xfcaller_to.remove( flow_of_interest );
		if ( f_xfcaller_to.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		Server from = network.findSplittingServer( to, f_xfcaller_to );
		alphas_xfcaller = ArrivalBound.computeArrivalBound( network, from, f_xfcaller, flow_of_interest );
		
		Flow f_representative = f_xfcaller_to.iterator().next();
		Path common_subpath = f_representative.getSubPath( from, link.getSource() );
		
		// Calculate the left-over service curves for ever server on the sub-path and convolve the cross-traffics arrival with it
		for ( Server s : common_subpath.getServers() )
		{
			Set<ServiceCurve> betas_lo_s;
			
			Set<Flow> f_xxfcaller_s = SetUtils.getDifference( network.getFlows( s ), f_xfcaller );
			f_xxfcaller_s.remove( flow_of_interest );

			Set<ArrivalCurve> alphas_xxfcaller_s = ArrivalBound.computeArrivalBound( network, s, f_xxfcaller_s, flow_of_interest );
			
			// Calculate the left-over service curve for this single server
			if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO
					|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL & s.useFifoMultiplexing() ) )
			{
				betas_lo_s = LeftOverService.fifoMux( s.getServiceCurve(), alphas_xxfcaller_s );
			}
			else
			{
				betas_lo_s = LeftOverService.arbMux( s.getServiceCurve(), alphas_xxfcaller_s );				
			}
			
			// The deconvolution of the two sets, arrival curves and service curves, respectively, takes care of all the possible combinations
			if( Configuration.useGamma() != GammaFlag.GLOBALLY_OFF  )
			{
				ServiceCurve gamma = common_subpath.getGamma();
				alphas_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alphas_xfcaller, gamma), betas_lo_s );
			}
			else
			{
				alphas_xfcaller = Deconvolution.deconvolve( alphas_xfcaller, betas_lo_s );
			}
			
			if( Configuration.useExtraGamma() != GammaFlag.GLOBALLY_OFF )
			{
				ServiceCurve extra_gamma = common_subpath.getExtraGamma();
				alphas_xfcaller = Convolution.convolve( alphas_xfcaller, extra_gamma );
			}
		}
		return alphas_xfcaller;
	}
}