/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2011 - 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 *
 */
public class PmooArrivalBound extends ArrivalBound {
	Network network;
	Flow flow_of_interest;
	
	@SuppressWarnings("unused")
	private PmooArrivalBound() {}
	
	public PmooArrivalBound( Network network, Flow flow_of_interest ) {
		this.network = network;
		this.flow_of_interest = flow_of_interest;
	}

	public Set<ArrivalCurve> computeArrivalBound( Server server ) throws Exception {
		return computeArrivalBound( server, network.getFlows( server ) );
	}
	
	public Set<ArrivalCurve> computeArrivalBound( Link link ) throws Exception {
		return computeArrivalBound( link, network.getFlows( link ) );
	}

	/**
	 * Computes the PMOO arrival bound for a set of
	 * <code>flows_to_bound</code>. The difference to the standard output
	 * bound method is that this method tries to compute tighter bounds by
	 * concatenating as many servers as possible using the PMOO approach. It
	 * does so by searching from <code>server</code> towards the sinks of the
	 * flows contained in <code>f_xfcaller</code> until it reaches the
	 * server where all these flows first meet each other (the
	 * "splitting point"). It then concatenates all servers between the
	 * splitting point (inclusive) and <code>server</code> (exclusive).
	 * 
	 * @param server
	 *            The server that all flows of interest flow into
	 * @param f_xfcaller
	 *            The set of flows of interest
	 * @return The PMOO arrival bounds
	 * @throws Exception 
	 */
	public Set<ArrivalCurve> computeArrivalBound( Server server, Set<Flow> f_xfcaller ) throws Exception {
		Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		
		f_xfcaller.remove( flow_of_interest );
		if ( f_xfcaller.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		Set<Flow> f_server = network.getFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		if ( f_xfcaller_server.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		Set<ArrivalCurve> arrival_bounds_link = new HashSet<ArrivalCurve>();
		Set<ArrivalCurve> arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
		
		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		f_xfcaller_sourceflows_server.remove( flow_of_interest );
		ArrivalCurve alpha_xfcaller_sourceflows_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );
		alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( alpha_xfcaller_sourceflows_server ) );
		
		// Get cross-traffic from each predecessor. Call per link in order to get splitting points
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		if ( !f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
			for ( Link in_l : network.getInLinks( server ) ) {
				arrival_bounds_link = new HashSet<ArrivalCurve>();
				arrival_bounds_link_merged = new HashSet<ArrivalCurve>();
				
				Set<Flow> f_xfcaller_in_l = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
				f_xfcaller_in_l.remove( flow_of_interest );
				
				if( f_xfcaller_in_l.isEmpty() ) { // Do not check links without flows to bound
					continue;
				}

				arrival_bounds_link = computeArrivalBound( in_l, f_xfcaller_in_l );
				
				if ( !alphas_xfcaller.isEmpty() ) {
					for ( ArrivalCurve arrival_bound_other_link : alphas_xfcaller ) {
						for ( ArrivalCurve arrival_bound_link : arrival_bounds_link ) {
							arrival_bound_link.beautify();
							arrival_bounds_link_merged.add( ArrivalCurve.add( arrival_bound_other_link, arrival_bound_link ) );
						}
					}
				} else {
					arrival_bounds_link_merged.addAll( arrival_bounds_link );
				}
				alphas_xfcaller = arrival_bounds_link_merged;
				
				f_xfcaller_server_bounded.addAll( f_xfcaller_in_l ); // Stop as soon as all flows are bounded
				if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
					break;
				}
			}
		}
		
		return alphas_xfcaller;
	}
	
	public Set<ArrivalCurve> computeArrivalBound( Link link, Set<Flow> f_xfcaller ) throws Exception {
		Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		if ( f_xfcaller.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		// Get the common sub-path of f_xfcaller flows crossing the given link
		Server to = link.getDest();
		Set<Flow> f_to = network.getFlows( to );
		Set<Flow> f_xfcaller_to = SetUtils.getIntersection( f_to, f_xfcaller );
		f_xfcaller_to.remove( flow_of_interest );
		if ( f_xfcaller_to.size() == 0 )
		{
			return alphas_xfcaller;
		}
		
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO 
				|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL & link.getSource().useFifoMultiplexing() ) )
		{
			throw new Exception( "PMOO arrival bounding is not available for FIFO multiplexing nodes" );
		}
		
		Server from = network.findSplittingServer( to, f_xfcaller_to );
		Flow f_representative = f_xfcaller_to.iterator().next();
		
		Path common_subpath = f_representative.getSubPath( from, link.getSource() );
		
		// Get cross-flows grouped as needed for the PMOO left-over service curve
		Set<Flow> f_xxfcaller_from_to = SetUtils.getUnion( f_xfcaller_to, Collections.singleton( flow_of_interest ) ); // Those flows at the server to should not be considered interference of themselves
		Map<Path,Set<Flow>> xtx_subpath_grouped = network.getFlowsPerSubPath( common_subpath, f_xxfcaller_from_to );
		
		// Derive the cross-flow substitutes with their arrival bound
		Set<List<Flow>> cross_flow_substitutes_set = new HashSet<List<Flow>>();
		cross_flow_substitutes_set.add( new LinkedList<Flow>() );
		Set<List<Flow>> cross_flow_substitutes_merged = new HashSet<List<Flow>>();
		
		for ( Map.Entry<Path,Set<Flow>> entry : xtx_subpath_grouped.entrySet() ) {
			cross_flow_substitutes_merged = new HashSet<List<Flow>>();
			
		// Create a single substitute flow 
	 		// Name the substitute flow
	 		String substitute_flow_alias = "subst_{";
	 		for( Flow f : entry.getValue() ) {
	 			substitute_flow_alias = substitute_flow_alias.concat( f.getAlias() + "," );
	 		}
	 		substitute_flow_alias = substitute_flow_alias.substring( 0, substitute_flow_alias.length()-1 );
	 		substitute_flow_alias = substitute_flow_alias.concat( "}" );

	 		// Derive the substitute flow's arrival bounds
			Set<ArrivalCurve> alphas_xf_group = ArrivalBound.computeArrivalBound( network, entry.getKey().getSource(), entry.getValue(), flow_of_interest ); // entry.getKey().getSource() because the path in the entry is the common subpath of all flows
	 		
			// Permutation with all the previously derived bounds
	 		for( ArrivalCurve alpha : alphas_xf_group ) {
	 			for( List<Flow> f_subst_list : cross_flow_substitutes_set ) {
 					List<Flow> tmp = new LinkedList<Flow>( f_subst_list );
 					tmp.add( Flow.createDummyFlow( substitute_flow_alias, alpha, entry.getKey() ) );
 					cross_flow_substitutes_merged.add( tmp );
	 			}
	 		}
			cross_flow_substitutes_set = new HashSet<List<Flow>>( cross_flow_substitutes_merged );
		}
				
		// Derive the left-over service curves
		Set<ServiceCurve> betas_loxfcaller_subpath = new HashSet<ServiceCurve>();
		ServiceCurve null_service = ServiceCurve.createNullService();
		
		for( List<Flow> xtx_substitutes : cross_flow_substitutes_set ) {
			ServiceCurve beta_loxfcaller_subpath = PmooAnalysis.getPmooServiceCurve( common_subpath, xtx_substitutes );
			
			if( !beta_loxfcaller_subpath.equals( null_service ) ) {
				betas_loxfcaller_subpath.add( beta_loxfcaller_subpath );
			}
		}
		
		// Check if there's any service left on this path. Signaled by at least one service curve in this set
		if( betas_loxfcaller_subpath.isEmpty() ) {
			// TODO Maybe throw a RuntimeException
			System.out.println( "No service left over" );
			System.exit( 0 );
		}
		
		// Get arrival bound at the splitting point:
		// We need to know the arrival bound of f_xfcaller at the server 'from', i.e., at the above sub-path's source
		// in order to deconvolve it with beta_loxfcaller_subpath to get the arrival bound of the sub-path
		// Note that flows f_xfcaller that originate in 'from' are covered by this call of computeArrivalBound
		Set<ArrivalCurve> alpha_xfcaller_from = ArrivalBound.computeArrivalBound( network, from, f_xfcaller, flow_of_interest );
		
		// Convolve to get the bound.
		// See "Improving Performance Bounds in Feed-Forward Networks by Paying Multiplexing Only Once", Lemma 2
		if( Configuration.useGamma() != GammaFlag.GLOBALLY_OFF  )
		{
			ServiceCurve gamma = common_subpath.getGamma();
			alphas_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_from, gamma), betas_loxfcaller_subpath );
		}
		else
		{
			alphas_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_from, betas_loxfcaller_subpath );
		}
		
		if( Configuration.useExtraGamma() != GammaFlag.GLOBALLY_OFF )
		{
			ServiceCurve extra_gamma = common_subpath.getExtraGamma();
			alphas_xfcaller = Convolution.convolve( alphas_xfcaller, extra_gamma );
		}
		
		return alphas_xfcaller;
	}
}
