/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.network;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.LinkedList;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.minplus.Convolution;

/**
 * 
 * @author Steffen Bondorf
 *
 */
public class Path {
	private LinkedList<Server> path_servers = new LinkedList<Server>();
	private LinkedList<Link> path_links = new LinkedList<Link>();

	public static Path createNullPath() {
		return new Path();
	}
	
	private Path() {};

	protected Path( List<Server> path_servers, List<Link> path_links ) {
		// Sanity check should have been done by the network
		this.path_servers = new LinkedList<Server>( path_servers );
		this.path_links = new LinkedList<Link>( path_links );
	}
	
	protected Path( Path path ) {
		this.path_servers = path.getServers();
		this.path_links = path.getLinks();
	}
	
	protected Path( Server single_hop ) {
		this.path_servers.add( single_hop );
	}
	
	public Server getSource(){
		return path_servers.get( 0 );
	}
	
	public Server getSink(){
		return path_servers.get( path_servers.size()-1 );
	}
	
	public int numServers(){
		return path_servers.size();
	}
	
	public int numLinks(){
		return path_links.size();
	}
	
	public LinkedList<Link> getLinks() {
		return new LinkedList<Link>( path_links );
	}
	
	public LinkedList<Server> getServers() {
		return new LinkedList<Server>( path_servers );
	}
	
    /**
     * @param from inclusive
     * @param to inclusive
     */
    public Path getSubPath( Server from, Server to ) throws Exception {
    	// All other sanity check should have been passed when this object was created
    	if( !path_servers.contains( from ) ) {
    		throw new Exception( "Source of the path servers are collected for is not crossed by this flow" );
    	}
    	if( !path_servers.contains( to ) ) {
    		throw new Exception( "Sink of the path servers are collected for is not crossed by this flow" );
    	}

    	if( from == to ) {
    		return new Path( new LinkedList<Server>( Collections.singleton( from ) ) , new LinkedList<Link>() );
    	}
    	
    	int from_index = path_servers.indexOf( from );
    	int to_index = path_servers.indexOf( to );
    	if ( from_index >= to_index ) {
    		throw new Exception( "Cannot create sub-path from " + from.toString() + " to " + to.toString() );
    	}
																	// subList: 'from' is inclusive but 'to' is exclusive
    	LinkedList<Server> subpath_servers = new LinkedList<Server>( path_servers.subList( from_index, to_index ) );
    	subpath_servers.add( to );
    	
    	List<Link> subpath_links = new LinkedList<Link>();
    	if ( subpath_servers.size() > 1 ) {
        	for ( Link l : path_links ) {
        		Server src_l = l.getSource();
        		Server snk_l = l.getDest();
        		if ( subpath_servers.contains( src_l ) && subpath_servers.contains( snk_l ) ) {
        			subpath_links.add( l );
        		}
        	}
    	}

    	return new Path( subpath_servers, subpath_links );
	}

 	public Link getPrecedingLink( Server s ) throws Exception {
 		for( Link l: path_links ) {
 			if ( l.getDest().equals( s ) ) {
 				return l;
 			}
 		}
 		throw new Exception( "No preceding link on the path found" );	
 	}

 	public Link getSucceedingLink( Server s ) throws Exception {
 		for( Link l: path_links ) {
 			if ( l.getSource().equals( s ) ) {
 				return l;
 			}
 		}
 		throw new Exception( "No succeeding link on the path found" );	
 	}

 	public Server getPrecedingServer( Server s ) throws Exception {
 		try{
 			return getPrecedingLink( s ).getSource();
 		} catch ( Exception e ) {
 			throw new Exception( "No preceding server on the path found" );
 		}
 	}

 	public Server getSucceedingServer( Server s ) throws Exception {
 		try{
 			return getSucceedingLink( s ).getDest();
 		} catch ( Exception e ) {
 			throw new Exception( "No succeeding server on the path found" );
 		}
 	}

	/**
	 * Just convolves the service curves on the path and returns the result.
	 * 
	 * @return The convolved curve
	 */
 	public ServiceCurve getServiceCurve()
	{
		Collection<Server> servers = getServers();
		return getServiceCurve( servers );
	}

	private ServiceCurve getServiceCurve( Collection<Server> servers )
	{
		ServiceCurve service_curve_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			service_curve_total = Convolution.convolve( service_curve_total, s.getServiceCurve() );
		}
		
		return service_curve_total;
	}
	
 	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given lnik path <code>path</code> that have the useGamma flag set.<br />
	 * If a server either has no maximum service curve set or useGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @return The convolved curve
	 */
	public ServiceCurve getGamma()
	{
		Collection<Server> servers = getServers();
		return getGamma( servers );
	}
	
	private ServiceCurve getGamma( Collection<Server> servers )
	{
		ServiceCurve gamma_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			gamma_total = Convolution.convolve( gamma_total, s.getGamma() );
		}
		
		return gamma_total;
	}

	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code> that have the useExtraGamma flag set.<br />
	 * If a server either has no maximum service curve set or useExtraGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @return The convolved curve
	 */
	public ServiceCurve getExtraGamma()
	{
		Collection<Server> servers = getServers();
		return getExtraGamma( servers );
	}

	private ServiceCurve getExtraGamma( Collection<Server> servers )
	{
		ServiceCurve extra_gamma_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			extra_gamma_total = Convolution.convolve( extra_gamma_total, s.getExtraGamma() );
		}	
		extra_gamma_total.removeLatency();
		
		return extra_gamma_total;
	}
	
	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code><br />
	 * If a server has no maximum service curve, calculations take place with 
	 * the default maximum service curve, i.e., the zero delay burst curve, 
	 * so the result will not be influenced.
	 * 
	 * @return The convolved curve
	 */
	public ServiceCurve getMaxServiceCurve()
	{
		Collection<Server> servers = getServers();
		return getMaxServiceCurve( servers );
	}

	private ServiceCurve getMaxServiceCurve( Collection<Server> servers )
	{
		ServiceCurve max_service_curve_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			max_service_curve_total = Convolution.convolve( max_service_curve_total, s.getMaxServiceCurve() );
		}
		
		return max_service_curve_total;
	}
 	
	@Override
	public String toString() {
		if ( path_links.isEmpty() && path_servers.isEmpty() ) {
			return "{}";
		}
		
		String str = "{";
		
		if ( path_links.isEmpty() && !path_servers.isEmpty() ) {
			return str = str.concat( "(" + path_servers.getFirst().toString() + ")}" );
		} else {
			for( Link l : path_links ) {
				str = str.concat( l.toString() + "," );
			}
			str = str.substring( 0, str.length()-1 );
			return str.concat( "}" );
		}
	}
}
