/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.tests;

import static org.junit.Assert.assertEquals;

import java.util.Collection;
import java.util.Set;

import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;
import unikl.disco.nc.Configuration.ArrivalBoundMethods;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.nc.TotalFlowAnalysis;
import unikl.disco.nc.SeparateFlowAnalysis;
import unikl.disco.nc.PmooAnalysis;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

@RunWith(value = Parameterized.class)
/**
 * 
 * @author Steffen Bondorf
 *
 */
public class Tandem_1SC_4Flows_1AC_1Path
{
	static final ServiceCurve service_curve = ServiceCurve.createRateLatency( 10, 10 );
	static final ArrivalCurve arrival_curve = ArrivalCurve.createTokenBucket( 2, 10 );
	static final ServiceCurve max_service_curve = ServiceCurve.createRateLatency( 15, 5 );

	static Network network;
	static Server s0, s1;
	static Flow f0, f1, f2, f3;
	
	@Parameters
	public static Collection<Object[]> data() {
		return FunctionalTests.createParameters();
	}
	 
	public Tandem_1SC_4Flows_1AC_1Path( Set<ArrivalBoundMethods> arrival_boundings, boolean iterative_ab, boolean remove_duplicates ) {
		FunctionalTests.printTestSettings( arrival_boundings, iterative_ab, remove_duplicates );
	}

	@BeforeClass
	static public void createNetwork()
	{
		network = new Network();

		s0 = network.addServer( service_curve, max_service_curve );
		s0.setUseGamma( false );
		s0.setUseExtraGamma( false );
		
		s1 = network.addServer( service_curve, max_service_curve );
		s1.setUseGamma( false );
		s1.setUseExtraGamma( false );

		try {
			network.addLink( s0, s1 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}

		try {	
			f0 = network.addFlow( arrival_curve, s0, s1 );
			f1 = network.addFlow( arrival_curve, s0, s1 );
			f2 = network.addFlow( arrival_curve, s0, s1 );
			f3 = network.addFlow( arrival_curve, s0, s1 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}
		
		Configuration.setPerformServiceCurveChecks( false );
		Configuration.setPerformArrivalCurveChecks( false );
		Configuration.setPerformFifoMultiplexingChecks( false );
		Configuration.setPerformDeconvolutionChecks( false );
	}
	
//--------------------Flow 0--------------------	
	@Test
	public void f0_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseFifoMultiplexing( true ); // default = false
		s1.setUseFifoMultiplexing( true ); // default = false
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f0 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA FIFO delay", new Num( 36 ), tf.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f0 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA ARB delay", new Num( 180 ), tf.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 34.5 ), sf.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 74 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 82.5 ), sf.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 170 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 60 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 125 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}

//--------------------Flow 1--------------------	
	@Test
	public void f1_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseFifoMultiplexing( true ); // default = false
		s1.setUseFifoMultiplexing( true ); // default = false
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f1 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA FIFO delay", new Num( 36 ), tf.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f1 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA ARB delay", new Num( 180 ), tf.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 34.5 ), sf.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 74 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 82.5 ), sf.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 170 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 60 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 125 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
//--------------------Flow 2--------------------	
	@Test
	public void f2_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseFifoMultiplexing( true ); // default = false
		s1.setUseFifoMultiplexing( true ); // default = false
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f2 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA FIFO delay", new Num( 36 ), tf.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f2 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA ARB delay", new Num( 180 ), tf.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 34.5 ), sf.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 74 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 82.5 ), sf.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 170 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 60 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 125 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}

//--------------------Flow 3--------------------	
	@Test
	public void f3_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseFifoMultiplexing( true ); // default = false
		s1.setUseFifoMultiplexing( true ); // default = false
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f3.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f3 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA FIFO delay", new Num( 36 ), tf.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f3_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f3.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		try {
			tf.performEnd2EndAnalysis( f3 );
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tf.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA ARB delay", new Num( 180 ), tf.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 200 ), tf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f3_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f3.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f3 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 34.5 ), sf.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 74 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f3_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f3.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		try {
			sf.performEnd2EndAnalysis( f3 );
			System.out.println( "e2e SFA SCs     : " + sf.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sf.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 82.5 ), sf.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 170 ), sf.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f3_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.SERVER_LOCAL );
		s0.setUseArbitraryMultiplexing( true ); // default = true
		s1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f3.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 60 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 125 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
}
