/*
 * This file is part of the Disco Deterministic Network Calculator v2.0.1 "Hydra".
 *
 * Copyright (C) 2013, 2014 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.tests;

import static org.junit.Assert.assertEquals;

import java.util.Collection;
import java.util.Set;

import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;
import unikl.disco.nc.TotalFlowAnalysis;
import unikl.disco.nc.SeparateFlowAnalysis;
import unikl.disco.nc.PmooAnalysis;
import unikl.disco.nc.Configuration.ArrivalBoundMethods;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

@RunWith(value = Parameterized.class)
/**
 * 
 * @author Steffen Bondorf
 *
 */
public class Tree_1SC_3Flows_1AC_3Paths
{
	static final ServiceCurve service_curve = ServiceCurve.createRateLatency( 20, 20 );
	static final ArrivalCurve arrival_curve = ArrivalCurve.createTokenBucket( 5, 25 );
	
	static Network network;
	static Server s0, s1, s2, s3, s4, s5, s6;
	static Flow f0, f1, f2;
	
	@Parameters
	public static Collection<Object[]> data() {
		return FunctionalTests.createParameters();
	}
	 
	public Tree_1SC_3Flows_1AC_3Paths( Set<ArrivalBoundMethods> arrival_boundings, boolean iterative_ab, boolean remove_duplicates ) {
		FunctionalTests.printTestSettings( arrival_boundings, iterative_ab, remove_duplicates );
	}
	
	@BeforeClass
	static public void createNetwork()
	{
		network = new Network();
		s0 = network.addServer( service_curve );
		s1 = network.addServer( service_curve );
		s2 = network.addServer( service_curve );
		s3 = network.addServer( service_curve );
		s4 = network.addServer( service_curve );
		s5 = network.addServer( service_curve );
		s6 = network.addServer( service_curve );

		try {
			network.addLink( s0, s1 );
			network.addLink( s1, s2 );
			network.addLink( s2, s5 );
			network.addLink( s3, s4 );
			network.addLink( s4, s5 );
			network.addLink( s5, s6 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}

		try {	
			f0 = network.addFlow( arrival_curve, s1, s6 );
			f1 = network.addFlow( arrival_curve, s0, s6 );
			f2 = network.addFlow( arrival_curve, s3, s6 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}
		
		Configuration.setPerformServiceCurveChecks( false );
		Configuration.setPerformArrivalCurveChecks( false );
		Configuration.setPerformFifoMultiplexingChecks( false );
		Configuration.setPerformDeconvolutionChecks( false );
	}

//--------------------Flow 0--------------------	
	@Test
	public void f0_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f0 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();
			
			assertEquals("TFA FIFO delay", new Num( 395, 2 ), tfa.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f0 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();

			assertEquals("TFA ARB delay", new Num( 660 ), tfa.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}

	@Test
	public void f0_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );

		try {
			sfa.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 165 ), sfa.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 1675, 2 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}

	@Test
	public void f0_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );

		try {
			sfa.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 1735, 6 ), sfa.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 4375, 3 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f0_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f0.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f0 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();

			assertEquals("PMOO ARB delay", new Num( 355, 2 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 900 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}

//--------------------Flow 1--------------------	
	@Test
	public void f1_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f1 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();
			
			assertEquals("TFA FIFO delay", new Num( 875, 4 ), tfa.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f1 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();
			
			assertEquals("TFA ARB delay", new Num( 2725, 4 ), tfa.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );

		try {
			sfa.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 165 ), sfa.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 1675, 2 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );
		
		try {
			sfa.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 1655, 6 ), sfa.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 4175, 3 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f1_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f1 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 375, 2  ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 950 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
//--------------------Flow 2--------------------	
	@Test
	public void f2_tfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f2 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();
			
			assertEquals("TFA FIFO delay", new Num( 180 ), tfa.getDelayBound() );
			assertEquals("TFA FIFO backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_tfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tTotal Flow Analysis (TFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
		
		try {
			tfa.performEnd2EndAnalysis( f2 );
			System.out.println( "delay bound     : " + tfa.getDelayBound() );
			System.out.println( "     per server : " + tfa.getServerDelayBoundMap().toString() );
			System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
			System.out.println( "     per server : " + tfa.getServerBacklogBoundMap().toString() );
			System.out.println( "alpha per server: " + tfa.getServerAlphasMap().toString() );
			System.out.println();
			
			assertEquals("TFA ARB delay", new Num( 1155, 2 ), tfa.getDelayBound() );
			assertEquals("TFA ARB backlog", new Num( 1375 ), tfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "TFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_sfa_fifoMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_FIFO );

		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tFIFO" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );

		try {
			sfa.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA FIFO delay", new Num( 295, 2 ), sfa.getDelayBound() );
			assertEquals("SFA FIFO backlog", new Num( 750 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_sfa_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tSeparate Flow Analysis (SFA)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f2.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );

		try {
			sfa.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
			System.out.println( "     per server : " + sfa.getServerLeftOverBetasMap().toString() );
			System.out.println( "xtx per server  : " + sfa.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + sfa.getDelayBound() );
			System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
			System.out.println();
			
			assertEquals("SFA ARB delay", new Num( 505, 2 ), sfa.getDelayBound() );
			assertEquals("SFA ARB backlog", new Num( 1275 ), sfa.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "SFA analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
	
	@Test
	public void f2_pmoo_arbMux()
	{
		Configuration.setMultiplexingDiscipline( MuxDiscipline.GLOBAL_ARBITRARY );
		
		System.out.println( "Analysis:\t\tPay Multiplexing Only Once (PMOO)" );
		System.out.println( "Multiplexing:\t\tArbitrary" );

		System.out.println( "Flow of interest:\t" + f1.toString() );
		System.out.println();
		
		System.out.println( "--- Results: ---" );
		PmooAnalysis pmoo = new PmooAnalysis( network );
		
		try {
			pmoo.performEnd2EndAnalysis( f2 );
			System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
			System.out.println( "xtx per server  : " + pmoo.getServerAlphasMap().toString() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
			
			assertEquals("PMOO ARB delay", new Num( 355, 2 ), pmoo.getDelayBound() );
			assertEquals("PMOO ARB backlog", new Num( 900 ), pmoo.getBacklogBound() );
		} catch (Exception e) {
			System.out.println( "PMOO analysis failed" );
			System.out.println( e.toString() );
			System.out.println();
		}
	}
}